#!/usr/bin/env python
"""
Script for logging data to FTP server.
"""

import os

try:
    import RPi.GPIO as GPIO
except ImportError:
    print 'WARNING: USING DUMMY GPIO MODULE. PLEASE INSTALL RPi.GPIO!'
    from dummy import GPIO
import struct
import time
import zlib
from ftpstream import FTPStream

PORTC = [21, 19, 15, 13, 11, 7, 5, 3]
PORTB7 = 23


class FTPLogger(object):

    def __init__(self, ftpstream, cycle_duration=300):
        self.ftpstream = ftpstream
        self.cycle_duration = cycle_duration
        self.num_channels = 2
        FTPLogger.setup()

    def __del__(self):
        print 'Cleaning up GPIO...'
        GPIO.cleanup()

    @staticmethod
    def _read_PORTC():
        v = 0
        for p in PORTC:
            v <<= 1
            if GPIO.input(p):
                v += 1
        return v

    @staticmethod
    def setup():
        GPIO.setmode(GPIO.BOARD)
        GPIO.setup(PORTC, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.setup(PORTB7, GPIO.IN, pull_up_down=GPIO.PUD_UP)

    def start(self):
        while True:
            start_time = None
            end_time = None
            scheduled_end_time = None
            samples = bytearray()
            while start_time is None or end_time < scheduled_end_time:
                GPIO.wait_for_edge(PORTB7, GPIO.FALLING)
                v0 = FTPLogger._read_PORTC()
                GPIO.wait_for_edge(PORTB7, GPIO.RISING)
                v1 = FTPLogger._read_PORTC()
                end_time = time.time()
                if start_time is None:
                    start_time = end_time
                    scheduled_end_time = start_time + self.cycle_duration
                # print '%5d %5d' % (v0, v1)
                samples.append(v0)
                samples.append(v1)
            print start_time, end_time, len(samples)
            data = bytearray(struct.pack('d', start_time) +
                             struct.pack('d', end_time) +
                             struct.pack('B', self.num_channels) +
                             struct.pack('i', len(samples) / self.num_channels))
            data.extend(samples)
            compressed = zlib.compress(bytes(data))
            # import binascii
            # print binascii.hexlify(compressed)
            fname = time.strftime('%Y-%m-%d-%H%M%S.log', time.gmtime())
            print 'Saving %s...' % fname
            self.ftpstream.enqueue(fname, compressed)


if __name__ == '__main__':
    ftpstream = FTPStream(host=os.environ['FTP_SERVER'],
                          user=os.environ['FTP_USER'],
                          password=os.environ['FTP_PASS'])
    ftpstream.start()
    logger = FTPLogger(ftpstream)
    logger.start()



"""
Dummy GPIO implementation for testing.
"""

import time
import math

class GPIO(object):
    """
    Dummy GPIO implementation.
    """
    INTERVAL = 0.05
    SIG_FREQ = 0.1
    SIG_AMPL = 100
    SIG_OFFSET = 50
    _signal = 0

    BOARD = 'BOARD'
    IN = 'IN'
    OUT = 'OUT'
    PUD_UP = 'PUD_UP'
    FALLING = 'FALLING'
    RISING = 'RISING'

    PORTC = [21, 19, 15, 13, 11, 7, 5, 3]

    @staticmethod
    def setmode(mode):
        print 'Dummy GPIO: setmode(%s)' % mode

    @staticmethod
    def setup(pin, inout, pull_up_down=None):
        print 'Dummy GPIO: setup(%s, %s, %s)' % (pin, inout, pull_up_down)

    @staticmethod
    def wait_for_edge(port, edge):
        GPIO._signal = (int(math.sin(3.14156 * time.time() * GPIO.SIG_FREQ) * GPIO.SIG_AMPL) + GPIO.SIG_OFFSET) & 0xFF;
        time.sleep(GPIO.INTERVAL)

    @staticmethod
    def cleanup():
        print 'Dummy GPIO: cleanup()'

    @staticmethod
    def input(pin):
        return '{0:08b}'.format(GPIO._signal)[GPIO.PORTC.index(pin)] == '1'



"""
FTP Stream library.
"""
import ftplib
import io
import threading
import time
import traceback


class FTPStream(object):
    """
    FTP upload for streaming data.
    """
    DEFAULT_POLL_INTERVAL = 10
    DEFAULT_TIMEOUT = 30

    def __init__(self, host, user, password, port=21, directory='.',
                 poll_interval=DEFAULT_POLL_INTERVAL, timeout=DEFAULT_TIMEOUT):
        """
        Constructor.

        Args:
            host (str): FTP server hostname
            user (str): FTP username
            password (str): FTP password
            port (int): FTP port
            directory (str): FTP (remote) directory
            poll_interval (int): poll interval, seconds
            timeout (int): connection timeout, seconds
        """
        self.host = host
        self.user = user
        self.password = password
        self.port = port or 21
        self.directory = directory or '.'
        self.poll_interval = poll_interval
        self.timeout = timeout
        self.queue = []
        self.running = False

    def _upload(self):
        while self.running or self.queue:
            while len(self.queue):
                print 'Queue length:', len(self.queue)
                try:
                    task = self.queue[0]
                    print 'Uploading %s...' % task['file_name']
                    self.ftp = ftplib.FTP(host=self.host, timeout=self.timeout)
                    self.ftp.login(user=self.user, passwd=self.password)
                    self.ftp.set_pasv(True)
                    self.ftp.cwd(self.directory)
                    self.ftp.storbinary('STOR %s' % task['file_name'], task['data'])
                    self.queue.pop(0)
                except:
                    traceback.print_exc()
                    time.sleep(self.poll_interval)
            time.sleep(self.poll_interval)

    def start(self):
        if self.running:
            return
        self.running = True
        upload_thread = threading.Thread(target=self._upload)
        upload_thread.start()

    def enqueue(self, file_name, data):
        """
        Enqueues a task.

        Args:
            file_name (str): target file name
            data (bytes): file data
        """
        self.queue.append({'file_name': file_name, 'data': io.BytesIO(data)})

    def stop_when_complete(self):
        self.running = False

    def stop(self):
        del self.queue[:]
        self.running = False



"""
Seismometer logfile reader.
"""

import argparse
import csv
import io
import os
import struct
import traceback
import zlib


class LogReader(object):
    """
    Seismometer logfile reader.
    """

    def __init__(self):
        """
        Constructor.
        """
        self.data = []
        self.num_channels = 0

    def clear(self):
        """
        Initializes the reader.
        """
        del self.data[:]
        self.num_channels = 0

    def unwrap(self):
        """
        Resolves the counter overflows.
        """
        self.sort()
        for c in range(1, self.num_channels+1):
            last = None
            correction = 0
            for i in range(len(self.data)):
                if i > 0:
                    dx = self.data[i][c] - last
                    if dx < -127:
                        correction += 256
                    elif dx > 127:
                        correction -= 256
                last = self.data[i][c]
                self.data[i][c] += correction

    def sort(self):
        """
        Sorts the samples by time.
        """
        self.data.sort(key=lambda item: item[0])

    def load_file(self, file_name):
        """
        Loads the data from a logfile.

        Args:
            file_name (str): path to a logfile
        """
        with open(file_name, 'rb') as handle:
            b = io.BytesIO(zlib.decompress(handle.read()))
            start_time, end_time, self.num_channels = struct.unpack('ddB', b.read(17))
            num_samples, = struct.unpack('i', b.read(4))
            # num_samples /= 2
            sample_interval = (end_time - start_time) / (num_samples - 1)
            t = start_time
            timestamps = []
            channel_data = []
            for c in range(self.num_channels):
                channel_data.append([])
            for sample in range(num_samples):
                try:
                    for c in range(self.num_channels):
                        v, = struct.unpack('B', b.read(1))
                        channel_data[c].append(v)
                    timestamps.append(t)
                    t += sample_interval
                except:
                    pass
        for i in range(len(timestamps)):
            record = [timestamps[i]]
            for c in range(self.num_channels):
                record.append(channel_data[c][i])
            self.data.append(record)

    def load_dir(self, directory):
        """
        Loads data from all files in a directory.

        Args:
            directory (str): directory with logfiles
        """
        for file_name in os.listdir(directory):
            if not file_name.lower().endswith('.log'):
                continue
            try:
                self.load_file(os.path.join(directory, file_name))
            except:
                traceback.print_exc()
        self.sort()

    def save_to_csv(self, file_name):
        """
        Saves data as CSV.

        Args:
            file_name (str): CSV file name (output)
        """
        self.sort()
        if not self.data:
            return
        with open(file_name, 'w') as handle:
            writer = csv.writer(handle)
            if self.data:
                writer.writerow(['TIMESTAMP'] + ['CHANNEL%d' % c for c in range(self.num_channels)])
                writer.writerows(self.data)

    def save_to_wav(self, file_name, channel=0, rate=128000):
        """
        Saves data as a WAV file.

        Args:
            file_name (str): a WAV file name (output)
        """
        import scipy.io.wavfile

        self.sort()
        if not self.data:
            return
        with open(file_name, 'w') as handle:
            scipy.io.wavfile.write(handle, data=self.data[:][channel], rate=rate)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--files', nargs='*', help='load data from those files')
    parser.add_argument('-d', '--directories', nargs='*', help='load data from those directories')
    parser.add_argument('-csv', '--save_to_csv', help='file name to save as CSV')
    args = parser.parse_args()
    logreader = LogReader()
    if args.directories:
        for directory in args.directories:
            logreader.load_dir(directory)
    if args.files:
        for file_name in args.files:
            logreader.load_file(file_name)
    logreader.unwrap()
    if args.save_to_csv:
        logreader.save_to_csv(args.save_to_csv)
